<?php
/**
 * Plugin Name: FOMO Pay Chinese Payment Solution
 * Plugin URI: https://www.fomopay.com/
 * Description: Accept Fomo Payments, orders are updated using APC.
 * Version: 2.0.0
 * Author: FOMO Pay
 * License: GPL-2.0+
 * License URI: https://www.gnu.org/licenses/gpl-2.0.txt
 * Tags: Payment, WeChat Pay, AliPay, UnionPay, FOMO Pay, NETs Pay, PAYNOW, FavePay, Online Payments, WooCommerce, WooCommerce Payment Gateway, WooCommerce Payment Gateway Extension
 */

if (!defined('WPINC')) {
    die;
}

add_action('plugins_loaded', 'woocommerce_fomo_init', 0);

function woocommerce_fomo_init()
{

    static $plugin;

    class WC_Fomo extends WC_Payment_Gateway
    {

        const FOMOPAY_STATUS_CREATED = 'CREATED';
        const FOMOPAY_STATUS_FAILED = 'FAIL';
        const FOMOPAY_STATUS_ERROR = 'ERROR';
        const FOMOPAY_STATUS_SUCCESS = 'SUCCESS';
        const FOMOPAY_STATUS_REFUND = 'REFUND';
        const FOMOPAY_STATUS_CLOSED = 'CLOSED';
        const FOMOPAY_STATUS_PENDING = 'PENDING';
        const FOMOPAY_STATUS_TIMEOUT = 'TIMEOUT';

        const GATEWAY_VERSION_IPG = 'ipg';
        const GATEWAY_VERSION_PGW = 'pgw';

        const FOMO_PAYMENT_STATUS = 'fomopay_payment_status';
        const FOMO_GATEWAY_VERSION = 'fomopay_gateway_version';
        const FOMO_TRANSACTION_ID = 'fomopay_transaction_id';
        const FOMO_ORDER_ID = 'fomopay_order_id';
        const FOMO_REFUND_HISTORY = 'fomopay_refund_history';

        private $PAYMENTS_SUPPORTED;
        private $email;
        private $api_version;
        private $api_type;
        private $api_key;
        private $sandbox_url;
        private $merchant;
        private $sub_merchant;
        //private $type;
        private $live_url;
        private $supported_currency;
        private $timeout;
        private $supported_channels;
        private $debugging;

        /**
         * he *Singleton* instance of this class
         *
         * @var WC_Fomo
         */
        private static $_instance;

        public static function get_instance()
        {
            if (self::$_instance === null) {
                self::$_instance = new self();
            }
            return self::$_instance;
        }

        public static function add_gateways($methods)
        {
            $methods[] = WC_Fomo::get_instance();
            return $methods;
        }

        public function register_rest_routers()
        {
            register_rest_route(
                'wc-fomo/v2',
                'callback_ipg',
                array(
                    'methods' => 'POST',
                    'callback' => array($this, 'handle_fomo_callback_ipg'),
                    'permission_callback' => '__return_true',
                )
            );
        }
        public function __construct()
        {

            $this->__init();

            // addition hook

            add_action('woocommerce_update_options_payment_gateways_' . $this->id, array($this, 'process_admin_options'));
            add_action('woocommerce_api_wc_fomo', array(&$this, 'handle_fomo_callback'));
            add_action('woocommerce_thankyou_' . $this->id, array(&$this, 'receipt_page_fomo'), 10, 1);
            add_filter('woocommerce_payment_gateways', array(&$this, 'add_gateways'));
            add_filter('woocommerce_available_payment_gateways', array(&$this, 'fomo_available'));

            add_action('wp_ajax_fomo_payment_status', array(&$this, 'handle_ajax_fomo_payment_status'));
            add_action('wp_ajax_nopriv_fomo_payment_status', array(&$this, 'handle_ajax_fomo_payment_status'));

            add_action('rest_api_init', [$this, 'register_rest_routers']);

            add_action('woocommerce_admin_order_data_after_billing_address', array($this, 'handle_adminpage_fomo_payment_status'), 10, 1);


        }

        private function __init()
        {

            global $woocommerce;
            $this->id = 'fomo';
            $this->icon = "";
            $this->has_fields = true;
            $this->PAYMENTS_SUPPORTED = [
                'wxpay' => 'Wechat Pay',
                'alipay' => 'Alipay',
                'spunionpay' => 'UnionPay',
                'creditcard' => 'Credit Card',
                'enetspat' => 'eNETS'
            ];
            $this->method_title = __('FOMO Pay', 'woocommerce');
            $this->supports = array('products', 'refunds');
            $this->method_description = __('FOMO Pay helps you to accept WeChat Pay, Alipay, UnionPay and credit cards in your WooCommerce store. Merchant ID, API key and supported currencies are needed to enable this solution.</br>
                <a href="https://accounts.fomopay.com/Signup">Click here to sign up your FOMO Pay account!</a>', 'woocommerce');

            // load the form fields.
            $this->init_form_fields();

            // load the settings.
            $this->init_settings();

            // load gateway configuration
            $this->enabled = $this->get_option('enabled');
            $this->title = $this->get_option('title');
            $this->description = $this->get_option('description');
            $this->email = $this->get_option('email');
            $this->api_version = $this->get_option('api_version');
            $this->api_type = $this->get_option('api_type');
            $this->api_key = htmlspecialchars_decode(htmlentities($this->get_option('api_key'), ENT_NOQUOTES, 'UTF-8', false), ENT_QUOTES);
            $this->sandbox_url = $this->get_option('sandbox_url');
            $this->merchant = trim($this->get_option('merchant'));
            $this->sub_merchant = trim($this->get_option('sub_merchant'));
            $this->live_url = $this->get_option('live_url');
            $this->supported_currency = $this->get_option('currency');
            $this->timeout = $this->get_option('timeout') ?
                $this->get_option('timeout') : 1800;
            //$this->supported_channels = $this->get_option('channels');
            $this->supported_channels = array_keys($this->PAYMENTS_SUPPORTED);
            $this->debugging = $this->get_option('debugging');
        }

        public function process_admin_options()
        {
            $post_data = $this->get_post_data();
            return parent::process_admin_options();
        }


        // check if the fomopay is available for use.
        public function fomo_available($available_gateways)
        {
            // disable fomopay option if store currency not in avaiable currencies list
            $wc_currency = strtolower(get_woocommerce_currency());
            if (!in_array($wc_currency, $this->supported_currency)) {
                if (isset($available_gateways['fomo'])) {
                    unset($available_gateways['fomo']);
                }
            }
            return $available_gateways;
        }

        // retreive payment result in receipt page
        function receipt_page_fomo($order_id)
        {
            global $woocommerce;

            $ajax_url = admin_url('admin-ajax.php');
            $order = new WC_Order($order_id);

            // if woocommerce version less than 3.0, use woocommerce_get_page_id;
            // else, use wc_get_page_id
            if (version_compare(WOOCOMMERCE_VERSION, '3.0', '<')) {
                $shop_page_url = get_permalink(woocommerce_get_page_id('shop'));
            } else {
                $shop_page_url = get_permalink(wc_get_page_id('shop'));
            }
            if ((get_post_meta($order_id, '_payment_method', true) == 'fomo')) {
                $transaction_id = $order->get_meta(self::FOMO_TRANSACTION_ID);
                if ($transaction_id == '' || $transaction_id == null) {
                    $transaction_id = get_post_meta($order_id, '_transaction_id', true);
                }
                echo '<div id="fomo-processing">';
                echo '<p>Thank you for using FOMO Pay</p>';
                echo '<p id="fomo-processing-text">We are processing your payment, please wait ...<img style="width:30px;height:30px;vertical-align: middle;" src="' . plugins_url('assets/images/loading.gif', __FILE__) . '"/></p>';
                echo '<p>Payment reference: <b>' . $transaction_id . '</b></p>';
                echo '</div>';
                echo '<div style="display: none;" id="fomo-payment-result">';
                echo '<p>Payment successful!</p>';
                echo '<p>Your payment reference is: <b>' . $transaction_id . '</b></p>';
                echo '<h4>' . __('Thanks for your order, we will send you confirmation email.<br><a href="' . esc_url($shop_page_url) . '" style="color:blue; text-decoration:none;">Continue shopping</a>', 'woocommerce') . '</h4>';
                echo '</div>';
            }
            ?>
            <script>
                (function () {
                    var counter = 0;
                    var ajaxUrl = '<?php echo $ajax_url ?>';
                    var timer = setInterval(function () {
                        jQuery.post(ajaxUrl, {
                            'action': 'fomo_payment_status',
                            'order_id': '<?php echo $order_id ?>'
                        }, function (status) {
                            console.log(status);
                            switch (status.trim()) {
                                case "created":
                                case "<?php echo self::FOMOPAY_STATUS_CREATED ?>":
                                    return;
                                case "success":
                                case "<?php echo self::FOMOPAY_STATUS_SUCCESS ?>":
                                    jQuery('#fomo-processing').css('display', 'none');
                                    jQuery('#fomo-payment-result').css('display', 'block');
                                    break;
                                case "fail":
                                case "<?php echo self::FOMOPAY_STATUS_FAILED ?>":
                                    jQuery('#fomo-processing').css('display', 'none');
                                    jQuery('#fomo-payment-result')
                                        .html('<p>Payment failed, please contact us.</p>' + '<p>Your payment reference is: <b>' + '<?php echo $transaction_id ?>' + '</b></p>')
                                        .css('color', 'red')
                                        .css('margin', '10px 0')
                                        .css('display', 'block')
                                    break;
                                default:
                                    jQuery('#fomo-processing').css('display', 'none');
                                    jQuery('#fomo-payment-result')
                                        .html('Unknown status, please contact us.')
                                        .css('color', 'red')
                                        .css('margin', '10px 0')
                                        .css('display', 'block')
                                    break;
                            }
                            clearInterval(timer);
                        })
                    }, 2000);
                })();
            </script>
            <?php
        }

        private function error_response($msg, $order = null)
        {
            $logger = wc_get_logger();
            $logger->info($msg, array('source' => 'fomopay-plugin'));
            if ($order) {
                $order->update_status('failed', __('Payment is failed.', 'woocommerce'));
            }
            http_response_code(403);
        }


        // initialize gateway configuration
        public function init_form_fields()
        {
            $channels = [];
            foreach ($this->PAYMENTS_SUPPORTED as $key => $value) {
                $channels[$key] = __($value, 'woocommerce');
            }

            $this->form_fields = array(
                'enabled' => array(
                    'title' => __('Enable/Disable', 'woocommerce'),
                    'type' => 'checkbox',
                    'label' => __('Enable FOMO Pay', 'woocommerce'),
                    'default' => 'no'
                ),
                'title' => array(
                    'title' => __('Method Title', 'woocommerce'),
                    'type' => 'text',
                    'description' => __('This is control title which display at user while checkout', 'woocommerce'),
                    'default' => __('FOMO Pay', 'woocommerce'),
                    'desc_tip' => true,
                ),
                'api_version' => array(
                    'title' => __('API Version', 'woocommerce'),
                    'type' => 'select',
                    'description' => __('Please Select API version', 'woocommerce'),
                    'default' => __('gateway', 'woocommerce'),
                    'options' => array(
                        'pgw' => __('Gateway', 'woocommerce'),
                        'ipg' => __('IPG', 'woocommerce')
                    ),
                ),
                'api_type' => array(
                    'title' => __('Sandbox ', 'woocommerce'),
                    'type' => 'checkbox',
                    'label' => __('Enable', 'woocommerce'),
                    'description' => __('Sandbox options are not supported in IPG mode', 'woocommerce'),
                    'default' => 'yes'
                ),
                'sandbox_url' => array(
                    'title' => __('Sandbox Url', 'woocommerce'),
                    'type' => 'text',
                    'description' => __('Please Enter Sandbox Url. ', 'woocommerce'),
                    'default' => __('https://gateway.fomopay.com/sandbox/pgw/v1', 'woocommerce'),
                    'desc_tip' => true,
                ),
                'live_url' => array(
                    'title' => __('Live Url', 'woocommerce'),
                    'type' => 'text',
                    'description' => __('Please Enter Live Url.', 'woocommerce'),
                    'default' => __('https://gateway.fomopay.com/pgw/v1', 'woocommerce'),
                    'desc_tip' => true,
                ),
                'api_key' => array(
                    'title' => __('API Key', 'woocommerce'),
                    'type' => 'text',
                    'description' => __('<a href="https://accounts.fomopay.com/Signup">Sign up and get your API key</a>', 'woocommerce'),
                    'default' => __('', 'woocommerce'),
                ),
                'merchant' => array(
                    'title' => __('Merchant', 'woocommerce'),
                    'type' => 'text',
                    'description' => __('<a href="https://accounts.fomopay.com/Signup">Sign up and get your Merchant Id</a>', 'woocommerce'),
                    'default' => __('', 'woocommerce'),
                ),
                'sub_merchant' => array(
                    'title' => __('Sub Merchant Id', 'woocommerce'),
                    'type' => 'text',
                    'description' => __('Please DO NOT set this field if not instructed, only supported in IPG mode.', 'woocommerce'),
                    'default' => __('', 'woocommerce'),
                ),
                // 'type' => array(
                //     'title' => __('Type', 'woocommerce'),
                //     'type' => 'checkbox',
                //     'label' => __('Sale', 'woocommerce'),
                //     'default' => 'yes'
                // ),
                'currency' => array(
                    'title' => __('Currency', 'woocommerce'),
                    'type' => 'multiselect',
                    'css' => 'width:400px;height:200px;',
                    'options' => array(
                        'sgd' => __('Singapore Dollar', 'woocommerce'),
                        'usd' => __('US Dollar', 'woocommerce'),
                        'myr' => __('Malaysia Ringgit', 'woocommerce'),
                        'eur' => __('Euro', 'woocommerce'),
                        'gbp' => __('British Pound', 'woocommerce'),
                        'cny' => __('Chinese Yuan', 'woocommerce'),
                        'hkd' => __('Hong Kong Dollar', 'woocommerce'),
                        'jpy' => __('Japanese Yen', 'woocommerce'),
                        'aud' => __('Australian Dollar', 'woocommerce'),
                    ),
                    'description' => __(' Select currencies your account supports.<br> Make sure your shop\'s currency is selected. </br> Hold Ctrl/Command for multiple selection', 'woocommerce'),
                    'default' => array(__('usd', 'woocommerce')),
                ),
                'description' => array(
                    'title' => __('Customer Message', 'woocommerce'),
                    'type' => 'textarea',
                    'css' => 'width:400px;',
                    'default' => 'Include WeChat Pay, Alipay, UnionPay and credit card payment',
                    'description' => __('The message which you want the customer see in checkout page.', 'woocommerce'),
                ),
                // 'channels' => array(
                //     'title' => __('Supported Payment Methods', 'woocommerce'),
                //     'type' => 'multiselect',
                //     'css' => 'width:400px;height:200px;',
                //     'options' => $channels,
                //     'description' => __(' Selected payment methods will be available in checkout page. </br> Hold Ctrl/Command for multiple selection', 'woocommerce'),
                //     'default' => array(__('wxpay', 'woocommerce')),
                // ),
                'timeout' => array(
                    'title' => __('Timeout', 'woocommerce'),
                    'type' => 'text',
                    'description' => __('maximum time for custom to complete payment', 'woocommerce'),
                    'default' => __('1800', 'woocommerce'),
                ),
                'debugging' => array(
                    'title' => __('Debug Mode', 'woocommerce'),
                    'type' => 'checkbox',
                    'label' => __('Enable Debug Mode', 'woocommerce'),
                    'description' => __('* Do not enable debug mode unless requested by FOMO Pay experts.', 'woocommerce'),
                    'default' => 'no'
                )
            );
        }

        // display fomopay configuration fields in admin page
        public function admin_options()
        {
            ?>
            <h3>
                <?php _e('FOMO Pay', 'woocommerce'); ?>
            </h3>
            <div>FOMO Pay helps you to accept WeChat Pay, Alipay, UnionPay and credit cards in your WooCommerce store. Merchant
                ID,
                API key and supported currencies are needed to enable this solution.</br>
                <a href="https://accounts.fomopay.com/Signup">Click here to sign up your FOMO Pay account</a>
            </div>
            <div id="poststuff">
                <div id="post-body" class="metabox-holder columns-2">
                    <div id="post-body-content">
                        <table class="form-table">
                            <?php $this->generate_settings_html(); ?>
                        </table>
                    </div>
                </div>
            </div>
            <div class="clear"></div>

            <script>
                jQuery(document).on('click', '.woocommerce-save-button', function () {
                    var api_key = jQuery('#woocommerce_fomo_api_key').val();
                    var new_api_key = escapeHtml(api_key);
                    jQuery('#woocommerce_fomo_api_key').val(new_api_key);
                    jQuery('#mainform').submit();
                });
                jQuery(document).ready(function ($) {
                    var api_version = $('#woocommerce_fomo_api_version');

                    var toggleFields = function () {
                        if (api_version.val() === 'pgw') {
                            $('#woocommerce_fomo_sub_merchant').prop('disabled', true);;
                            $('#woocommerce_fomo_api_type, #woocommerce_fomo_sandbox_url').prop('disabled', false);
                            $('#woocommerce_fomo_sandbox_url, #woocommerce_fomo_sandbox_url').prop('disabled', false);
                        } else {
                            $('#woocommerce_fomo_sub_merchant').prop('disabled', false);
                            $('#woocommerce_fomo_api_type, #woocommerce_fomo_sandbox_url').prop('disabled', true);
                            $('#woocommerce_fomo_sandbox_url, #woocommerce_fomo_sandbox_url').prop('disabled', true);
                        }
                    };

                    toggleFields();

                    api_version.change(toggleFields);
                });
                function escapeHtml(text) {
                    'use strict';
                    return text.replace(/[\"&<>]/g, function (a) {
                        return { '"': '&quot;', '&': '&amp;', '<': '&lt;', '>': '&gt;' }[a];
                    });
                }
            </script>
            <?php
        }

        // override default checkout page
        public function payment_fields()
        {
            ?>
            <div>
                <p>FOMO Pay supports:</p>
                <div id="fomo-icon-list">
                    <img title="Wechat Pay" id="fomo-wxpay" class="payment_icon"
                        src="<?php echo plugins_url('assets/images/wechatpay.jpg', __FILE__) ?>" />
                    <img title="Alipay" id="fomo-alipay" class="payment_icon"
                        src="<?php echo plugins_url('assets/images/alipay.jpg', __FILE__) ?>" />
                    <img title="UnionPay" id="fomo-unionpay" class="payment_icon"
                        src="<?php echo plugins_url('assets/images/unionpay.png', __FILE__) ?>" />
                    <img title="eNETS" id="fomo-enets" class="payment_icon"
                        src="<?php echo plugins_url('assets/images/enets.png', __FILE__) ?>" />
                    <img title="Creditcard" id="fomo-creditcard" class="payment_icon"
                        src="<?php echo plugins_url('assets/images/creditcard.png', __FILE__) ?>" />
                    <img title="Diners" id="fomo-diners" class="payment_icon"
                        src="<?php echo plugins_url('assets/images/diners.png', __FILE__) ?>" />
                    <img title="Discover" id="fomo-discover" class="payment_icon"
                        src="<?php echo plugins_url('assets/images/discover.png', __FILE__) ?>" />
                </div>

                <script>
                    var enabledChannels = <?php echo json_encode($this->supported_channels) ?>;
                    enabledChannels.forEach(function (ch) {
                        if (ch == 'creditcard') {
                            jQuery('#fomo-diners').css('display', 'block');
                            jQuery('#fomo-discover').css('display', 'block');
                        }
                        jQuery('#fomo-' + ch).css('display', 'block');
                    })
                </script>
                <style>
                    #fomo-icon-list {
                        display: flex;
                        flex-wrap: wrap;
                        flex-direction: row;
                        justify-content: space-around;
                        margin: 6px;
                    }

                    .payment_icon {
                        max-height: 2.6em !important;
                        margin: 3px 6px !important;
                        display: none;
                    }
                </style>
                <p>
                    <?php echo $this->description ?>
                </p>
            </div>
            <?php
        }

        // create order
        public function process_payment($order_id)
        {
            $version = $this->api_version;
            if ($version == self::GATEWAY_VERSION_PGW) {
                return $this->process_payment_pgw($order_id);
            } else {
                return $this->process_payment_ipg($order_id);
            }
        }

        private function process_payment_pgw($order_id)
        {
            global $woocommerce;

            $order = new WC_Order($order_id);

            $logger = wc_get_logger();
            $logger->info('Order: ' . $order_id . ' is created.', array('source' => 'fomopay-plugin'));

            $url = $this->get_gateway_url();
            $merchant = $this->merchant;
            $shared_key = htmlspecialchars_decode($this->api_key, HTML_ENTITIES);

            // create payment parameters
            $status = 'created';
            $transaction_id = $this->get_fomo_transaction_id($order_id);
            $order->update_meta_data(self::FOMO_PAYMENT_STATUS, $status);
            $order->update_meta_data(self::FOMO_TRANSACTION_ID, $transaction_id);
            // if wc version < 3.0, use get_order_currency()
            // else, use get_currency();
            if (method_exists($order, 'get_currency')) {
                $order_currency = strtolower($order->get_currency());
            } else {
                $order_currency = strtolower($order->get_order_currency());
            }
            $nonce = $this->make_nonce();
            $cart_total = $order->get_total();
            $my_type = "sale";
            $default_return_url = $this->get_return_url($order);
            $return_url = str_replace("checkout-2", "checkout", $default_return_url);
            $callback_url = add_query_arg('wc-api', 'wc_fomo', home_url('/'));
            $timeout = 1800;
            if ($this->timeout != '') {
                $timeout = $this->timeout;
            }
            $description = preg_replace(array('/\n/', '/\r/', '/\r\n/', '/\t/'), ' ', $order->customer_note);
            $price = sprintf("%.2f", $cart_total);

            $upstreams = [];
            foreach ($this->supported_channels as $key) {
                if ($key == 'wxpay') {
                    array_push($upstreams, 'wxpay,wxpayqr');
                    continue;
                }

                array_push($upstreams, $key);
            }
            $upstreams = join(',', $upstreams);

            $payload = array(
                'merchant' => $merchant,
                'price' => $price,
                'description' => $description,
                'transaction' => $transaction_id,
                'return_url' => $return_url,
                'callback_url' => $callback_url,
                'currency_code' => $order_currency,
                'type' => $my_type,
                'timeout' => $timeout,
                'nonce' => $nonce
            );
            if (count($this->supported_channels) > 0) {
                $payload['upstreams'] = $upstreams;
            }
            $signature = $this->sign_gateway($payload);

            $payload['signature'] = $signature;
            $payload['url'] = $url;

            if ($this->debugging == 'yes') {
                $shared_key_masked = $shared_key;
                $key_length = strlen($shared_key);
                if ($key_length > 10) {
                    $shared_key_masked = substr_replace($shared_key_masked, "***LENGTH:$key_length***", 5, -5);
                }
                $this->remoteLog(array_merge($payload, [
                    'shared_key' => $shared_key_masked
                ]));
            }

            $order->update_meta_data(self::FOMO_PAYMENT_STATUS, self::FOMOPAY_STATUS_CREATED);
            $order->update_meta_data(self::FOMO_GATEWAY_VERSION, self::GATEWAY_VERSION_PGW);
            $order->save();

            $queryParams = http_build_query($payload);
            return array(
                'result' => 'success',
                'redirect' => plugins_url('form.php', __FILE__) . '?' . $queryParams
            );
        }

        private function process_payment_ipg($order_id)
        {

            $order = new WC_Order($order_id);
            $logger = wc_get_logger();
            $logger->info('Order: ' . $order_id . ' is created.', array('source' => 'fomopay-plugin'));

            $url = $this->get_gateway_url();
            $mode = 'HOSTED';
            $sub_mid = $this->sub_merchant;
            $subject = 'Order #' . $order_id;
            $description = $subject;
            $amount = sprintf("%.2f", $order->get_total());
            if (method_exists($order, 'get_currency')) {
                $currency = strtoupper($order->get_currency());
            } else {
                $currency = strtoupper($order->get_order_currency());
            }
            $order_no = $this->get_fomo_transaction_id($order_id);
            $notify_url = home_url('/wp-json/wc-fomo/v2/callback_ipg');
            $return_url = str_replace("checkout-2", "checkout", $this->get_return_url($order));
            $back_url = $return_url;

            $data = array(
                "mode" => $mode,
                "orderNo" => $order_no,
                "subject" => $subject,
                "description" => $description,
                "amount" => $amount,
                "currencyCode" => $currency,
                "notifyUrl" => $notify_url,
                "returnUrl" => $return_url,
                "backUrl" => $back_url,
            );

            if ($sub_mid != null && $sub_mid != '') {
                $data['subMid'] = $sub_mid;
            }

            $args = array(
                'url' => $url,
                'method' => 'POST',
                'headers' => array(
                    'Content-Type' => 'application/json',
                    'Authorization' => $this->get_ipg_auth_header()
                ),
                'body' => wp_json_encode($data)
            );

            //$logger->info('IPG Request: ' . $args['body'], array('source' => 'fomopay-plugin'));

            $response = wp_remote_post($url, $args);
            $response_code = wp_remote_retrieve_response_code($response);
            $response_body = wp_remote_retrieve_body($response);
            $response_data = json_decode($response_body, true);

            //$logger->info('IPG Response: ' . $response_body, array('source' => 'fomopay-plugin'));

            $payment_status = self::FOMOPAY_STATUS_FAILED;
            $redirect_url = $return_url;

            if ($response_code >= 200 && $response_code <= 300 && isset($response_data['url'])) {
                $payment_status = $response_data['status'];
                $redirect_url = $response_data['url'];
                $order->update_meta_data(self::FOMO_ORDER_ID, $response_data['id']);
                $order->update_meta_data(self::FOMO_PAYMENT_STATUS, $payment_status);
                $order->update_meta_data(self::FOMO_GATEWAY_VERSION, self::GATEWAY_VERSION_IPG);
                $order->update_meta_data(self::FOMO_TRANSACTION_ID, $order_no);
            }

            $order->save();

            return array(
                'result' => 'success',
                'redirect' => $redirect_url,
            );
        }

        // create refund
        public function process_refund($order_id, $amount = null, $reason = '')
        {
            $logger = wc_get_logger();
            $order = new WC_Order($order_id);

            // Refund is only supported when the gateway version equals the current version.

            if ($amount < 0.01) {
                return new WP_Error('IllegalAmount', "refund amount must be at least 0.01.");
            }
            if (method_exists($order, 'get_currency')) {
                $currency = $order->get_currency();
            } else {
                $currency = $order->get_order_currency();
            }
            $transaction_id = $order->get_meta(self::FOMO_TRANSACTION_ID);

            if (!$currency || !$transaction_id) {
                return new WP_Error(
                    'IllegalOrder',
                    "order currency or transaction id is empty."
                );
            }

            $order_gateway_version = $order->get_meta(self::FOMO_GATEWAY_VERSION);
            if ($order_gateway_version != $this->api_version) {
                return new WP_Error('IllegalGatewayVersion', "order api version is not equal to current option, please contact the administrator.");
            }

            if ($this->api_version == self::GATEWAY_VERSION_IPG) {
                return $this->process_refund_ipg($order, $amount, $reason);
            } else {
                return $this->process_refund_pgw($order, $amount, $reason);
            }
        }

        private function process_refund_ipg($order, $amount = null, $reason = '')
        {
            $logger = wc_get_logger();
            if ($amount < 0.01) {
                return new WP_Error('IllegalAmount', "refund amount must be at least 0.01.");
            }
            if (method_exists($order, 'get_currency')) {
                $currency = $order->get_currency();
            } else {
                $currency = $order->get_order_currency();
            }
            $transaction_id = $order->get_meta(self::FOMO_TRANSACTION_ID);
            $fomo_payment_id = $order->get_meta(self::FOMO_ORDER_ID);

            if (!$currency || !$transaction_id) {
                return new WP_Error(
                    'IllegalOrder',
                    "order currency or transaction id is empty."
                );
            }

            $refund_history = $order->get_meta(self::FOMO_REFUND_HISTORY, true);
            if (!$refund_history) {
                $refund_history = '{}';
            }
            $refund_history = json_decode($refund_history, true);
            $refund_transaction_no = $this->get_fomo_refund_id($transaction_id, $refund_history);
            $refund_meta = array(
                'transaction_no' => $refund_transaction_no,
                'transaction_id' => '',
                'amount' => $amount,
                'reason' => $reason,
                'status' => self::FOMOPAY_STATUS_CREATED,
                'created_at' => time(),
            );

            $refund_payload = array(
                'type' => 'REFUND',
                'transactionNo' => $refund_meta['transaction_no'],
                'amount' => $amount,
                'currencyCode' => strtoupper($currency),
                'subject' => $reason,
            );

            // create sub transactions under the fomopay payment id
            $refund_url = $this->get_gateway_url()
                . '/' . $fomo_payment_id . '/transactions/';
            $response = wp_remote_post(
                $refund_url,
                array(
                    'method' => 'POST',
                    'headers' => array(
                        'Content-Type' => 'application/json',
                        'Authorization' => $this->get_ipg_auth_header()
                    ),
                    'body' => wp_json_encode($refund_payload)
                )
            );
            $logger->info('IPG Refund Request URL: ' . $refund_url, array('source' => 'fomopay-plugin'));
            //$logger->info('IPG Refund Request: ' . wp_json_encode($refund_payload), array('source' => 'fomopay-plugin'));
            //$logger->info('IPG Refund Response: ' . wp_json_encode($response), array('source' => 'fomopay-plugin'));
            if (is_wp_error($response)) {
                $error_message = $response->get_error_message();
                $logger->error("Something went wrong: $error_message");
                return new WP_Error(
                    'Error',
                    'Refund Failed, please contact FOMO Pay for help, code 1001.'
                );
            }

            $response_code = wp_remote_retrieve_response_code($response);

            if ($response_code != 200 && $response_code != 201) {
                return new WP_Error(
                    'Error',
                    'Refund Failed, create refund error, please contact FOMO Pay for help, code 1002.'
                );
            }

            $response_body = wp_remote_retrieve_body($response);
            $response_data = json_decode($response_body, true);
            $logger->info('IPG Refund Response: ' . $response_body, array('source' => 'fomopay-plugin'));

            if (isset($response_data['id'])) {
                $refund_transaction_id = $response_data['id'];
                $refund_status = $response_data['status'];
                $refund_meta['transaction_id'] = $refund_transaction_id;
                $refund_meta['status'] = $refund_status;
                $refund_history[$refund_transaction_no] = $refund_meta;
                $order->update_meta_data(self::FOMO_REFUND_HISTORY, json_encode($refund_history));
                $order->add_order_note(
                    sprintf(
                        __('Refunds in process, amount %1$s - Sub-transaction ID: %2$s </br>Please contact Fomo Pay to continue the transaction.', 'woocommerce'),
                        $amount,
                        $refund_transaction_id
                    )
                );
                $order->save();
            }

            return $this->update_ipg_refund_status($order, $refund_transaction_id, $refund_transaction_no, $fomo_payment_id);

        }

        private function process_refund_pgw($order, $amount = null, $reason = '')
        {
            $logger = wc_get_logger();
            if ($amount < 0.01) {
                return new WP_Error('IllegalAmount', "refund amount must be at least 0.01.");
            }
            if (method_exists($order, 'get_currency')) {
                $currency = $order->get_currency();
            } else {
                $currency = $order->get_order_currency();
            }
            $transaction_id = $order->get_meta(self::FOMO_TRANSACTION_ID);
            $shared_key = htmlspecialchars_decode($this->api_key, HTML_ENTITIES);

            if (!$shared_key) {
                return false;
            }

            $refund_payload = array(
                'currency_code' => strtolower($currency),
                'merchant' => $this->merchant,
                'nonce' => $this->make_nonce(),
                'price' => $amount,
                'transaction' => $transaction_id,
            );

            $signature = $this->sign_gateway($refund_payload);
            $refund_payload['signature'] = $signature;
            $refund_url = 'https://gateway.fomopay.com/pgw/v1/refund';
            $response = wp_remote_post(
                $refund_url,
                array(
                    'method' => 'POST',
                    'timeout' => 45,
                    'redirection' => 5,
                    'httpversion' => '1.0',
                    'blocking' => true,
                    'headers' => array(),
                    'body' => $refund_payload
                )
            );

            if (is_wp_error($response)) {
                $error_message = $response->get_error_message();
                $logger->error("Something went wrong: $error_message");
                return new WP_Error('Error', 'Refund Failed, please contact FOMO Pay for help, code 1001.');
            }

            try {
                $response_payloads = json_decode($response['body'], true);
                if ($response_payloads['result'] >= 400) {
                    if ($response_payloads['result'] === 409) {
                        $logger->warning($response['body']);
                        return new WP_Error('Confilict', 'Currently only support refund once, if you want to do another refund for same order, please contact Fomo Pay.');
                    } else {
                        $logger->error($response['body']);
                        return new WP_Error('Error', $response_payloads['reason']);
                    }
                } else {
                    return true;
                }
            } catch (Exception $e) {
                return new WP_Error('Error', 'Refund Failed, please contact FOMO Pay for help, code: 1002.');
            }
        }

        // send log to FOMO remote logger
        private function remoteLog($content)
        {
            $remote_url = 'http://log.webplugin.fomopay.net/woocommerce';
            $response = wp_remote_post(
                $remote_url,
                array(
                    'method' => 'POST',
                    'timeout' => 45,
                    'redirection' => 5,
                    'httpversion' => '1.0',
                    'blocking' => false,
                    'headers' => array(
                        'Content-Type' => 'application/json'
                    ),
                    'body' => json_encode($content)
                )
            );
        }

        // handle fomopay gateway callback
        public function handle_fomo_callback()
        {
            global $woocommerce;
            @ob_clean();

            $logger = wc_get_logger();

            // remove double splashes in string
            $_REQUEST = stripslashes_deep($_REQUEST);

            // check callback parameter
            if (
                !isset($_REQUEST['transaction']) ||
                !isset($_REQUEST['payment_id']) ||
                !isset($_REQUEST['result']) ||
                !isset($_REQUEST['upstream']) ||
                !isset($_REQUEST['nonce']) ||
                !isset($_REQUEST['signature'])
            ) {
                $this->error_response("Invalid parameter.");
            }

            $transaction_id = sanitize_text_field($_REQUEST['transaction']);
            $payment_id = sanitize_text_field($_REQUEST['payment_id']);
            $result = sanitize_text_field($_REQUEST['result']);
            $upstream = sanitize_text_field($_REQUEST['upstream']);
            $nonce = sanitize_text_field($_REQUEST['nonce']);
            $signature = sanitize_text_field($_REQUEST['signature']);

            $to_sign = [];
            $to_sign['transaction'] = $transaction_id;
            $to_sign['payment_id'] = $payment_id;
            $to_sign['result'] = $result;
            $to_sign['nonce'] = $nonce;
            $to_sign['upstream'] = $upstream;
            $signature_calculated = $this->sign_gateway($to_sign);
            if ($signature != $signature_calculated) {
                $this->error_response("Invalid signature");
            }

            $order_id = $this->get_wc_order_id($transaction_id);

            $logger->info('Order: ' . $order_id . ' has incoming callback.', array('source' => 'fomopay-plugin'));

            try {
                $order = new WC_Order($order_id);
                if (!$order) {
                    $this->error_response("Order was lost during payment.");
                }
            } catch (Exception $e) {
                $logger->error("Order not found.");
            }

            if (!is_numeric($result)) {
                $this->error_response("Invalid result.");
            }

            if ((get_post_meta($order_id, '_payment_method', true) == 'fomo')) {

                $payment_status = self::FOMOPAY_STATUS_FAILED;
                if ($result == 0) {
                    $payment_status = self::FOMOPAY_STATUS_SUCCESS;
                }

                $this->handle_payment_status($order, $transaction_id, $payment_status, $payment_id, $upstream);

            } else {
                $this->error_response("This order is not processed by FOMO Pay.", $order);
            }
        }
        public function handle_fomo_callback_ipg(WP_REST_Request $request)
        {

            $logger = wc_get_logger();

            $body = $request->get_body();
            //$logger->info('IPG callback body: ' . $body, array('source' => 'fomopay-plugin'));

            $json_data = json_decode($body, true);
            $fomo_payment_id = $json_data['orderId'];
            $transaction_id = $json_data['orderNo'];

            $payment_data = $this->process_ipg_query($fomo_payment_id);

            //$logger->info('IPG callback payment data: ' . json_encode($payment_data), array('source' => 'fomopay-plugin'));

            $order_id = $this->get_wc_order_id($transaction_id);
            $order = new WC_Order($order_id);

            if (!$order) {
                $this->error_response("Order was lost during payment.");
            }

            $payment_status = $payment_data['status'];
            $upstream = $payment_data['sourceOfFund'];
            $this->handle_payment_status($order, $transaction_id, $payment_status, $fomo_payment_id, $upstream);

            return new WP_REST_Response('OK', 200);
        }


        public function handle_ajax_fomo_payment_status()
        {
            $logger = wc_get_logger();
            $logger->info('Chcking status for order ' . $_POST['order_id'], array('source' => 'fomopay-plugin'));

            $order_id = $_POST['order_id'];
            $order = new WC_Order($order_id);
            $status = $order->get_meta(self::FOMO_PAYMENT_STATUS, true);
            if (!$status) {
                $status = get_post_meta($order_id, '_status', true);
            }

            $logger->info('Status to be passed to receiving page: ' . $status, array('source' => 'fomopay-plugin'));

            echo $status;
            wp_die();
        }

        /**
         * @param WC_Order $order
         * @param string $payment_status
         * @return void
         */
        private function handle_payment_status($order, $transaction_id, $payment_status, $payment_id, $upstream = null)
        {
            $logger = wc_get_logger();
            $order_id = $order->get_id();


            // try to fetch current status from order meta, if not found, try post meta
            $current_status = $order->get_meta(self::FOMO_PAYMENT_STATUS, true);
            if (!$current_status) {
                $current_status = get_post_meta($order_id, '_status', true);
            }

            $timeout = $this->timeout;
            $created_at = $order->get_date_created()->getTimestamp();
            if (
                $payment_status == self::FOMOPAY_STATUS_CREATED
                && $created_at + $timeout < time()
            ) {
                $payment_status = self::FOMOPAY_STATUS_TIMEOUT;
            }

            if ($current_status == $payment_status) {
                return;
            }

            $logger->info('Order: ' . $order_id . ' is updated to ' . $payment_status . ' through callback, current status is ' . $current_status, array('source' => 'fomopay-plugin'));

            $message = '';
            $message = $message . 'Transaction ID: ' . $transaction_id . '<br/>';
            $message = $message . 'Payment ID: ' . $payment_id . '<br/>';
            if (isset($upstream)) {
                $message = $message . 'Upstream: ' . $upstream . '<br/>';
            }
            $message = $message . 'Payment Status: ' . $payment_status . '<br/>';

            $order->add_order_note($message);
            $order->update_meta_data(self::FOMO_PAYMENT_STATUS, $payment_status);
            $order->save();

            try {
                switch ($payment_status) {
                    case self::FOMOPAY_STATUS_SUCCESS: {
                            $order->payment_complete();
                            break;
                        }
                    case self::FOMOPAY_STATUS_ERROR:
                    case self::FOMOPAY_STATUS_FAILED: {
                            $order->update_status('failed', __('Payment has been declined by FOMO Pay.', 'woocommerce'));
                            break;
                        }
                    case self::FOMOPAY_STATUS_CLOSED: {
                            $order->update_status('cancelled', __('Payment has been cancelled.', 'woocommerce'));
                            break;
                        }
                    case self::FOMOPAY_STATUS_TIMEOUT: {
                            $order->update_status('on-hold', __('Payment has been timeout.', 'woocommerce'));
                            break;
                        }
                    case self::FOMOPAY_STATUS_REFUND: {
                            break;
                        }
                }
            } catch (Exception $e) {
                $logger->error("Error while updating order status.");
            }
        }

        private function update_ipg_refund_status($order, $refund_transaction_id, $refund_transaction_no, $fomo_payment_id, $queryflag = false)
        {


            $logger = wc_get_logger();
            if (isset($refund_transaction_id) && $refund_transaction_id != '') {
                $retries = 3;
                if ($queryflag)
                    $retries = 1;
                $refund_status = 'CREATED';
                while ($retries--) {
                    $query_response = $this->process_ipg_query($fomo_payment_id, $refund_transaction_id);

                    $logger->info('IPG Refund Query Response: ' . json_encode($query_response), array('source' => 'fomopay-plugin'));

                    if ($query_response['type'] != 'REFUND') {
                        return new WP_Error('Error', 'Refund Failed, transaction not found, please contact FOMO Pay for help.');
                    }
                    $refund_status = $query_response['status'];
                    if ($refund_status != 'CREATED') {
                        break;
                    }
                    sleep(2);
                }

                // Get the latest refund history to avoid save duplicate data
                $refund_history = $order->get_meta(self::FOMO_REFUND_HISTORY);
                if ($refund_history != '') {
                    $refund_history = json_decode($refund_history, true);
                } else {
                    $refund_history = array();
                }
                if (!isset($refund_history[$refund_transaction_no])) {
                    return new WP_Error('Error', 'Refund Failed, transaction not found, please contact FOMO Pay for help.');
                }

                $refund_history[$refund_transaction_no]['status'] = $refund_status;
                $order->update_meta_data(self::FOMO_REFUND_HISTORY, json_encode($refund_history));
                $created_at = $query_response['createdAt'];
                if ($refund_status == 'CREATED' && $created_at + $this->timeout < time()) {
                    $refund_status = self::FOMOPAY_STATUS_TIMEOUT;
                }
                $order->save();

                if ($refund_status == 'SUCCESS') {
                    $order->add_order_note(
                        sprintf(
                            __('Refund success, ID: %1$s', 'woocommerce'),
                            $refund_transaction_id
                        )
                    );
                    if ($queryflag) {
                        wc_create_refund(
                            array(
                                'amount' => $refund_history[$refund_transaction_no]['amount'],
                                'reason' => $refund_history[$refund_transaction_no]['reason'],
                                'order_id' => $order->get_id(),
                            )
                        );
                    }
                    return true;
                } else if ($refund_status == 'FAIL') {
                    $order->add_order_note(
                        sprintf(
                            __('Refund failed, ID: %1$s', 'woocommerce'),
                            $refund_transaction_id
                        )
                    );
                    return new WP_Error(
                        'Error',
                        'Refund Failed, please contact FOMO Pay for help, code 1003.'
                    );
                } else if ($refund_status == 'CREATED' && !$queryflag) {
                    $order->add_order_note(
                        sprintf(
                            __('Refund transaction created, ID: %1$s', 'woocommerce'),
                            $refund_transaction_id
                        )
                    );
                    return new WP_Error(
                        'Processing',
                        'Refund transaction created, please refresh the page to check the latest status.'
                    );
                } else if ($refund_status == 'TIMEOUT') {
                    $order->add_order_note(
                        sprintf(
                            __('Refund Transaction timeout, ID: %1$s', 'woocommerce'),
                            $refund_transaction_id
                        )
                    );
                    $order->update_status('on-hold', 'Order refund transaction has timed out.');
                    return new WP_Error(
                        'Error',
                        'Refund timedout, please contact FOMO Pay for help, code 1004.'
                    );
                }
            }
            return new WP_Error(
                'Error',
                'Refund Failed, please contact FOMO Pay for help, code 1002.'
            );
        }

        private function admin_display_ipg_meta($order)
        {

            // 1. display order meta
            // if payment status is 'CREATED', 
            // trigger order query logic.

            $transaction_id = $order->get_meta(self::FOMO_TRANSACTION_ID);
            $payment_status = $order->get_meta(self::FOMO_PAYMENT_STATUS);
            $fomo_payment_id = $order->get_meta(self::FOMO_ORDER_ID);

            if ($payment_status == self::FOMOPAY_STATUS_CREATED) {
                $payment_data = $this->process_ipg_query($fomo_payment_id);
                $payment_status = $payment_data['status'];
                $this->handle_payment_status($order, $transaction_id, $payment_status, $fomo_payment_id);
            }

            echo '<p><strong>Payment Status:</strong> ' . $payment_status . '</p>';
            echo '<p><strong>Transaction ID:</strong> ' . $transaction_id . '</p>';
            echo '<p><strong>IPG Payment ID:</strong> ' . $fomo_payment_id . '</p>';

            // 2. display refund history.
            // if refund status is 'CREATED', 
            // trigger refund query logic.

            $refund_history = $order->get_meta(self::FOMO_REFUND_HISTORY);
            if ($refund_history != '') {
                $refund_history = json_decode($refund_history, true);
            } else {
                $refund_history = array();
            }

            $index = 0;

            if (count($refund_history) > 0) {
                foreach ($refund_history as $k => $refund_meta) {
                    $transaction_no = $refund_meta['transaction_no'];
                    $refund_transaction_id = $refund_meta['transaction_id'];
                    $amount = $refund_meta['amount'];
                    $reason = $refund_meta['reason'];
                    $refund_status = $refund_meta['status'];
                    if ($refund_status == 'CREATED') {
                        $this->update_ipg_refund_status($order, $refund_transaction_id, $transaction_no, $fomo_payment_id, true);
                    }
                }
            }

            if (count($refund_history) > 0) {
                $refund_history = $order->get_meta(self::FOMO_REFUND_HISTORY);
                if ($refund_history != '') {
                    $refund_history = json_decode($refund_history, true);
                } else {
                    $refund_history = array();
                }
                ?>
                <p><strong>Refund History:</strong></p>
                <table class="fomopay_refund_history">
                    <thead>
                        <tr>
                            <th>#</th>
                            <th>Refund ID</th>
                            <th>Amount</th>
                            <th>Reason</th>
                            <th>Status</th>
                        </tr>
                    </thead>
                    <tbody>
                        <?php
                        foreach ($refund_history as $k => $refund_meta) {
                            $index++;
                            $transaction_no = $refund_meta['transaction_no'];
                            $transaction_id = $refund_meta['transaction_id'];
                            $amount = $refund_meta['amount'];
                            $reason = $refund_meta['reason'];
                            $refund_status = $refund_meta['status'];
                            $created_at = $refund_meta['created_at'];
                            echo '<tr>';
                            echo '<td>' . $index . '</td>';
                            echo '<td>' . $transaction_id . '</td>';
                            echo '<td>' . $amount . '</td>';
                            echo '<td>' . $reason . '</td>';
                            echo '<td>' . $refund_status . '</td>';
                        }
                        ?>
                    </tbody>
                </table>
                <?php
            }

        }

        private function admin_display_pgw_meta($order)
        {
            $order_id = $order->get_id();
            $transaction_id = $order->get_meta(self::FOMO_TRANSACTION_ID);
            if ($transaction_id == '' || $transaction_id == null) {
                $transaction_id = get_post_meta($order_id, '_transaction_id', true);
            }
            $payment_status = $order->get_meta(self::FOMO_PAYMENT_STATUS);
            if (!$payment_status) {
                $payment_status = get_post_meta($order_id, '_status', true);
            }
            echo '<p><strong>Payment Status:</strong> ' . $payment_status . '</p>';
            echo '<p><strong>Transaction ID:</strong> ' . $transaction_id . '</p>';

        }

        public function handle_adminpage_fomo_payment_status($order)
        {
            ?>
            <div class="fomopay_order_info">
                <h3>FOMO Pay Payment Information</h3>
                <?php

                $gateway_version = $order->get_meta(self::FOMO_GATEWAY_VERSION);
                if ($gateway_version == self::GATEWAY_VERSION_IPG) {
                    echo '<p><strong>API Version:</strong>IPG</p>';
                    $this->admin_display_ipg_meta($order);
                } else {
                    echo '<p><strong>API Version:</strong>GATEWAY</p>';
                    $this->admin_display_pgw_meta($order);
                }
                ?>
            </div>
            <?php
        }

        private function get_gateway_url()
        {
            if ($this->api_type == 'yes' && $this->api_version = self::GATEWAY_VERSION_PGW) {
                return $this->sandbox_url;
            }
            return $this->live_url;
        }

        private function process_ipg_query($fomo_order_id, $transaction_id = null)
        {
            $query_url = $this->get_gateway_url() . '/' . $fomo_order_id;
            if (isset($transaction_id)) {
                $query_url = $query_url . '/transactions/' . $transaction_id;
            }
            $args = array(
                'headers' => array(
                    'Authorization' => $this->get_ipg_auth_header()
                )
            );
            $logger = wc_get_logger();
            $logger->info('IPG query url: ' . $query_url, array('source' => 'fomopay-plugin'));
            $response = wp_remote_get($query_url, $args);
            $response_data = json_decode($response['body'], true);
            return $response_data;
        }

        private function get_ipg_auth_header()
        {
            $username = $this->merchant;
            $password = $this->api_key;
            $authorizationType = "Basic";
            $basicCredentials = base64_encode($username . ':' . $password);
            $authHeader = $authorizationType . " " . $basicCredentials;
            return $authHeader;
        }

        private function get_fomo_transaction_id($order_id)
        {
            return 'TRACXV1410' . date('dmyhis') . '_' . $order_id;
        }

        private function get_fomo_refund_id($transaction_id, $refund_history)
        {

            $len = count($refund_history);
            $refund_id = $transaction_id . '_REFUND_' . date('dmyhis') . '_' . ($len + 1);
            return $refund_id;
        }

        private function get_wc_order_id($transaction_id)
        {
            if (!preg_match('/^[a-zA-Z0-9\-\_]*\_[a-zA-Z0-9]+$/', $transaction_id)) {
                $this->error_response("Invalid order_id.");
            }
            $order_explosive = explode('_', $transaction_id);
            return absint($order_explosive[1]);
        }

        private function make_nonce()
        {
            $timestamp = time();
            $rand = random_int(0, 99999999);
            $rand_pad = sprintf("%08d", $rand);
            return $timestamp . $rand_pad;
        }

        private function sign_gateway($payload)
        {
            $shared_key = htmlspecialchars_decode($this->api_key, HTML_ENTITIES);
            ;
            ksort($payload);

            $has_nonce = false;
            $buff = [];
            foreach ($payload as $key => $value) {
                if ($key == 'nonce') {
                    $has_nonce = true;
                }
                if ($key == 'signature') {
                    continue;
                }
                $buff[] = ($key . '=' . $value);
            }

            if (!$has_nonce) {
                // refuse to sign because empty nonce may introduce security issues.
                throw new Exception('Cannot sign payload without nonce.');
            }
            $buff[] = ('shared_key=' . $shared_key);
            $params = implode("&", $buff);

            return hash("sha256", $params);
        }
    }

    if (!isset($plugin)) {
        $plugin = WC_Fomo::get_instance();
    }

    return $plugin;
}